/* Sample code Lissajous3D_2 by Rudy Rucker, CS Dept, SJSU, Sept 25, 2003.
This code is public domain and can be freely reused. */

#include <windows.h>   // use as needed for your system
#include <gl/Gl.h>  //Must include this before vectortransformation
#include <gl/glut.h>
#include <time.h> //For the clock() function used by sleep() for animation.
#include <math.h> //For cos and sine functions
#include "randomizer.h"
#include "vectortransformation.h"

cRandomizer RR;
GLdouble ZOOMRATIO = 0.8;
GLdouble startwindowwidth = 640.0;
GLdouble aspect = 4.0/3.0;
GLdouble xradius = 4.0;
GLdouble yradius = 3.0;
GLdouble zradius = 3.0;
cVector eye(2.0*xradius, -4.0*yradius, 9.0*zradius);
double zoom = 1.0;
GLdouble bobradius = 0.4;
GLdouble ribbonhalfwidth = 1.0;
GLdouble ribbondz = 0.1;
GLdouble TIMEPERFRAME = 0.02; //Time in seconds per frame
bool animateflag = true;
bool threedflag = true;
GLdouble xfreq = 5.0;
GLdouble yfreq = 3.0;
GLdouble zfreq = 7.0;
GLdouble startt = 0.0;
GLdouble targetdt = 0.15;
GLdouble dt = 0.01;
int tsteps = 120;
//<<<<<<<<<<<<<<<<<<<<<<< myInit >>>>>>>>>>>>>>>>>>>>

void setProjectionMatrix(double zoom)
{
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	if (!threedflag) //Use a 2D ortho projection.
		glOrtho(-zoom *xradius, zoom *xradius, -zoom *yradius, zoom *yradius,
			 -zradius, zradius); 
		/* Arguments: left, right, bottom, top, near, far.  The near and far are
		relative to a viewer located at the origin, so here we normally
		want a negative near and a positive far. We zoom only in the z and y,
		as zooming in the z might clip unnecessariy. */
	else //Use a 3D perspective projection
	{
		double boxradius = sqrt(xradius*xradius + yradius*yradius + zradius*zradius);
		double eyetoorigin = eye.magnitude();
		gluPerspective(zoom*45.0, aspect, eyetoorigin-boxradius, eyetoorigin+boxradius);
		/* Arguments: view angle in degrees, width-to-height aspect ratio, 
		near, far.  The aspect should match the viewport dx/dy.  The near
		and far should be both positive, as here we think of the eye as
		standing back from the scene. */
		gluLookAt(eye.x(), eye.y(), eye.z(), 0.0, 0.0, 0.0, 0.0, 0.0, 1.0);
		/* Arguments: 3 eye coords, 3 center coords, 3 up direction components.
		This puts the eye at the orgin of a viewer coordinate system and puts
		the "center" on the negative z axis. */
	}
}

void myInit(void)
{
/* We call myInit once at startup. */
//    glClearColor(0.0, 0.0, 0.0, 0.0);       // set black background color
    glClearColor(1.0, 1.0, 1.0, 0.0);       // set white background color
 	glPointSize(4.0);		       // a dot is 4 by 4 pixels
	setProjectionMatrix(1.0);
	glShadeModel(GL_FLAT);
	glEnable(GL_DEPTH_TEST); //In case I want to use 3D
}

//<<<<<<<<<<<<<<<<<<<<<<<< utilities >>>>>>>>>>>>>>>>>
void sleep(double secondstowait)
{
/* This makes the program spin in a loop for an interval of secondstowait.
It can make functionality a bit awkward for long secondstowait, as you aren't
polling the keyboard and mouse while sleep spins, which can make the
program feel dead.*/
	long waitticks = secondstowait * CLOCKS_PER_SEC; 
		//CLOCKS_PER_SEC is typically 30, it's a system thing.
	long goal = waitticks + clock();
	while( goal > clock() )
      ; //Spin a few thousand times.
}
//<<<<<<<<<<<<<<<<<<<<<<<< drawSphere >>>>>>>>>>>>>>>>>
void drawSphere(GLdouble x, GLdouble y, GLdouble z, GLdouble radius)
{
	glMatrixMode(GL_MODELVIEW); //Ready to move objects into place
	glPushMatrix();
	glTranslated(x, y, z);
	glutSolidSphere(radius, 16, 16); //Second two argumetns are slices and stacks.
	glColor3f(0.0, 0.0, 0.0); 
	glLineWidth(1.0);
	glutWireSphere(1.1*radius, 16, 16); //Second two argumetns are slices and stacks.
	glPopMatrix();
}

//<<<<<<<<<<<<<<<<<<<<<<< drawLiss >>>>>>>>>>>>>>>>>

void fixdt()
{
	int maxfreq = __max(xfreq, yfreq);
	maxfreq = __max(maxfreq, zfreq);
	dt = targetdt / maxfreq;
}

void drawLiss() //Uses the global afreq, bfreq, startt, dt, and tsteps params.
{
	fixdt();
	double t = startt;
	double oldt = startt - dt;
	glColor3f(0.0, 0.0, 0.0);

	cVector oldpoint = cVector(xradius * cos(xfreq * oldt),
		yradius * sin(yfreq * oldt), zradius * cos(zfreq * oldt));
	cVector newpoint = 	cVector(xradius * cos(xfreq * t),
		yradius * sin(yfreq * t), zradius * cos(zfreq * t));
	cVector oldtangent(0,1,0);
	cVector newtangent = (newpoint - oldpoint).normalize();
	cVector normal;
	cVector oldnormal(1,0,0);
	cVector binormal(0,0,1);
		
		for (int i = 0; i < tsteps; i++) 
		{
			oldpoint = newpoint;
			oldtangent = newtangent;
			oldnormal = normal;
			t += dt;
			newpoint = cVector(xradius * cos(xfreq * t), yradius * sin(yfreq * t), 
				zradius * cos(zfreq * t));
			newtangent = (newpoint - oldpoint).normalize();
			binormal = newtangent * oldtangent;
			normal = binormal * newtangent;
			glLineWidth(2.0);
			glColor3f(1.0, 0.0, 0.0); 
			glBegin(GL_LINE_LOOP);
				(newpoint + ribbonhalfwidth * normal).glVertex();
				(oldpoint + ribbonhalfwidth * oldnormal).glVertex();
				(oldpoint - ribbonhalfwidth * oldnormal).glVertex();
				(newpoint - ribbonhalfwidth * normal).glVertex();
			glEnd();
			glColor3f(0.0, 0.0, 1.0); 
			glBegin(GL_LINE_LOOP);
			(newpoint + ribbonhalfwidth * normal -
					ribbondz*binormal).glVertex(); //a
				(newpoint - ribbonhalfwidth * normal -
					ribbondz*binormal).glVertex(); //d
				(oldpoint - ribbonhalfwidth * oldnormal-
					ribbondz*binormal).glVertex(); //c
				(oldpoint + ribbonhalfwidth * oldnormal-
					ribbondz*binormal).glVertex(); //b
			glEnd();
			glColor3f(1.0, 0.9, 0.9); 
			glBegin(GL_QUADS); //CCW
				(newpoint + ribbonhalfwidth * normal).glVertex(); //a
				(oldpoint + ribbonhalfwidth * oldnormal).glVertex(); //b
				(oldpoint - ribbonhalfwidth * oldnormal).glVertex(); //c
				(newpoint - ribbonhalfwidth * normal).glVertex(); //d
			glEnd();
			glColor3f(0.9, 0.9, 1.0); 
			glBegin(GL_QUADS); //CW
				(newpoint + ribbonhalfwidth * normal -
					ribbondz*binormal).glVertex(); //a
				(newpoint - ribbonhalfwidth * normal -
					ribbondz*binormal).glVertex(); //d
				(oldpoint - ribbonhalfwidth * oldnormal-
					ribbondz*binormal).glVertex(); //c
				(oldpoint + ribbonhalfwidth * oldnormal-
					ribbondz*binormal).glVertex(); //b
			glEnd();
		}
	//Back up to the last vertex and draw a ball there.
	t-=dt;
	glColor3f(1.0, 1.0, 0.0); //yellow
	drawSphere(xradius * cos(xfreq * t), yradius * sin(yfreq * t),
		zradius * cos(zfreq * t), bobradius);
}  
//<<<<<<<<<<<<<<<<<<<<<<<< myDisplay >>>>>>>>>>>>>>>>>
void myDisplay(void)
{
/* myDisplay is called once at start up and is called again whenever
the graphics window is first covered and then uncovered, making a
redraw necessary. */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);     // clear the screen 
//Draw 3D box around the world.
	glColor3f(0.8, 0.8, 0.8); //Light gray
	glLineWidth(1.0);
	glBegin(GL_LINE_LOOP); //loz face
		glVertex3d(-xradius, -yradius, -zradius);
		glVertex3d(xradius, -yradius, -zradius);
		glVertex3d(xradius, yradius, -zradius);
		glVertex3d(-xradius, yradius, -zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //hizface
		glVertex3d(-xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, zradius);
		glVertex3d(xradius, yradius, zradius);
		glVertex3d(-xradius, yradius, zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //loyface
		glVertex3d(-xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, zradius);
		glVertex3d(xradius, -yradius, -zradius);
		glVertex3d(-xradius, -yradius, -zradius);
	glEnd();
	glBegin(GL_LINE_LOOP); //hiyface
		glVertex3d(-xradius, yradius, zradius);
		glVertex3d(xradius, yradius, zradius);
		glVertex3d(xradius, yradius, -zradius);
		glVertex3d(-xradius, yradius, -zradius);
	glEnd();
	//Draw coordinate axes in Red, Blue, Yellow
	glLineWidth(3.0);
	glBegin(GL_LINES);
		glColor3f(1.0, 0.0, 0.0); //X axis red
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(1.0, 0.0, 0.0);
		glColor3f(0.0, 0.0, 1.0); //Y axis blue
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 1.0, 0.0);
		glColor3f(1.0, 1.0, 0.0); //Z axis yellow
		glVertex3d(0.0, 0.0, 0.0);
		glVertex3d(0.0, 0.0, 1.0);
	glEnd();
	//Draw Lissjous curve
	drawLiss();//Uses the global afreq, bfreq, startt, dt, and tsteps params.
	startt += dt; //Scoot the starting point forward.
	glFlush();	// Rush all output to display.  More important with single buffer
	glutSwapBuffers();		//Doule buffer: flip.
}
//<<<<<<<<<<<<<<<<<<<<<<<< myReshape >>>>>>>>>>>>>>>>>

void myReshape(int width, int height)
{
	aspect = double(width)/double(height);
	setProjectionMatrix(zoom);
	glViewport(0, 0, width, height);
}
//<<<<<<<<<<<<<<<<<<<<<<<< myAnimate >>>>>>>>>>>>>>>>>
void myAnimate(void)
{
/* myAnimate is called over and over whenever glut is idle (not busy
processing a key or mouse message). To keep this from happening too
often we put in a delay, that is, we use a spin loop to make myAnimate
take at least TIMEPERFRAME seconds. */
	if (!animateflag)
		return;
	long waitticks = TIMEPERFRAME * CLOCKS_PER_SEC; 
		//CLOCKS_PER_SEC is typically 30 "ticks" per second, as I recall.
	long goal = waitticks + clock();
		//clock() returns a long integer that counts something like "ticks".
	myDisplay();
	while( goal > clock() )
      ; //Spin if there's unused time.
}
//<<<<<<<<<<<<<<<<<<<<<<<< myKeyboard >>>>>>>>>>>>>>>>>
void myKeyboard(unsigned char key, int mx, int my)
{
/* key is the ASCII code of the key, such as 'a', 'A', 'b', 'B', and
so on.  Note that case matters.  mx and my are the current mouse
position. */
	switch(key)
	{
		case ' ': //Space
			myDisplay(); //redraw 
			break;
		case 'a':
			animateflag ^= TRUE;
			break;
		case 'q':
			exit(-1); //Terminate
			break;
		case 'r':
			yfreq = xfreq = zfreq = 2 + RR.random(2, 15);
				/* Make xfreq between 2 and 7.  I 
				want yfreq to be different, so I
				start by making it same, then loop
				till it's not. */
			while (yfreq == xfreq || zfreq == xfreq || zfreq == yfreq)
			{
				yfreq = 2 + RR.random(2, 15);
				zfreq = 2 + RR.random(2, 15);
			}
			break;
		case 'z':
			zoom *= ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
		case 'Z':
			zoom *= 1.0/ZOOMRATIO;
			setProjectionMatrix(zoom);
			break;
		case '2':
			threedflag = false;
			setProjectionMatrix(zoom);
			break;
		case '3':
			threedflag = true;
			setProjectionMatrix(zoom);
			break;
	}
}
//<<<<<<<<<<<<<<<<<<<<<<<< myMouse >>>>>>>>>>>>>>>>>
void myMouse(int button, int state, int mx, int my)
{
/* Gets called when a button is clicked.  The arguments are:
button is GLUT_LEFT_BUTTON or GLUT_RIGHT_BUTTON,
state is GLUT_DOWN or GLUT_UP, 
and mx and my are window pixel coordinates, origin at window upper left. */
}
//<<<<<<<<<<<<<<<<<<<<<<<< myKeyboard >>>>>>>>>>>>>>>>>
void myMovedMouse(int mx, int my)
{
/*Gets called whenever the mouse moves, and mx and my are the new window
pixel coordinates, origin at window upper left. */
}
//<<<<<<<<<<<<<<<<<<<<<<<< main >>>>>>>>>>>>>>>>>>>>>>
void main(int argc, char** argv)
{
	glutInit(&argc, argv);          // initialize the toolkit
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB ); // set display mode
	glutInitWindowSize(startwindowwidth, int(startwindowwidth/aspect));     // set window size
	glutInitWindowPosition(100, 150); // set window position on screen
	glutCreateWindow("Lissajous3D, Ver. 2, Sept 25, 2003, Rudy Rucker"); // open the screen window
	glutDisplayFunc(myDisplay);     // register redraw function
	glutReshapeFunc(myReshape);
	glutKeyboardFunc(myKeyboard);	// register keyboard function.
	glutMouseFunc(myMouse);			// register mouse click function.
	glutMotionFunc(myMovedMouse);   // register mouse move function.
	glutIdleFunc(myAnimate);		// register idle function to call.
	myInit();                   
	glutMainLoop(); 		     // go into a perpetual loop
}
